class OBJECT
inherit
	TICKABLE
		rename
			die as old_die
		redefine
			blast,
			shot,
			player_touch
		end
	HAS_PROGRAM
	DIRECTION
	KNOWS_MASTER
	KNOWS_RAND
creation make, make_for_world
feature
	id : STRING
	id_lower : STRING
		-- always use lowercase for comparison purposes
		-- we need to preserve the original id string because
		-- sometimes we print it on the screen

	walk_aim : INTEGER
	locked : BOOLEAN

	read_init_rest(header : OBJECT_HEADER) is
	local
		i : INTEGER
	do
		put_chr(header.p1)
		current_offset := header.ins + 1
		if current_offset < 1 then
			current_offset := 1
		end
		read_program(header.data)
		-- check first line for ID
		!!id.make(0)
		!!id_lower.make(0)
		if program.count > 0 then
			if program.item(1) = '@' then
				i := 2
				from
				until program.item(i) = '%R'
				loop
					id.add_last(program @ i)
					id_lower.add_last((program @ i).to_lower)
					i := i + 1
				end
			end
		end
		square.map.object_list.add_last(Current)
	ensure
		id_lower /= Void
		id /= Void
	end

	time_up : BOOLEAN
	old_offset : INTEGER
	go_forward : BOOLEAN

	cycle_tick is
	do
		execute_one_tick
	end

	goto_count : INTEGER
	goto_max : INTEGER is 32

	execute_one_tick is
		--execute one tick's worth from current line
	local
		i : INTEGER
		sq : SQUARE
		d : INTEGER
		label : STRING
		s : STRING
		lines : ARRAY[STRING]
		finished : BOOLEAN
	do
		goto_count := 0
		dungeon_master.put_current_object(Current)
		time_up := false
		!!lines.make(1, 0)
		from
		until current_offset > program.count or time_up
		loop
			io.put_string("Object (")
			io.put_integer(square.location.x)
			io.put_string(", ")
			io.put_integer(square.location.y)
			io.put_string(") [")
			io.put_string(id)
			if is_bound then
				io.put_string(" (B)")
			end
			io.put_string("] ")
			io.put_integer(current_offset)
			io.put_string("/")
			io.put_integer(program.count)
			io.put_string(": ")
			print_current_line
			io.put_string("%N")
			go_forward := true
			old_offset := current_offset
			inspect program.item(current_offset)
			when '/' then --blocking move
				current_offset := current_offset + 1
				command_go
				if not go_forward then
					current_offset := old_offset
				end
				time_up := true
			when '?' then --try move
				current_offset := current_offset + 1
				command_try
				time_up := true
			when '%'' then --comment
				next_line
			when '@' then --object id, should be parsed already
				next_line
			when ':' then --label
				next_line
			when '#' then --command
				current_offset := current_offset + 1
				process_command
			when '%R' then --blank line
				if not lines.empty then
					lines.add_last("")
				end
				current_offset := current_offset + 1
			else
				--scroll

				!!s.make(50)
				from
				until program.item(current_offset) = '%R'
				loop
					s.add_last(program.item(current_offset))
					current_offset := current_offset + 1
				end
				lines.add_last(s)
				current_offset := current_offset + 1
			end
		end

		if walk_aim >= min_dir and then walk_aim <= max_dir then
			if not try_walk(walk_aim) then
				goto("thud")
			end
		end

		if not lines.empty then
			dungeon_master.zzt.show_scroll(lines)
		end

	end

	process_command is
	local
		command : STRING
	do
		command := next_word
		if program @ current_offset = ':' then
			--undocumented syntax for sending messages
			send_message(command)
			next_line
		elseif command_table.has(command) then
			command_table.at(command).execute
			if go_forward then
				next_line
			else
				current_offset := old_offset
			end
		elseif find_label(command) <= program.count then
			--undocumented syntax for sending messages
			goto(command)
		else
			io.put_string("unimplemented command: ")
			io.put_string(command)
			io.put_string("%N")
			next_line
			time_up := true
		end
	end

	print_current_line is
	local
		i : INTEGER
	do
		from i := current_offset
		until i > program.count or else program @ i = '%R'
		loop
			io.put_character(program @ i)
			i := i + 1
		end
	end

	--better ways of doing this...
	command_give is
	local
		s : STRING
		i : INTEGER
		p : PLAYER
	do
		p := dungeon_master.zzt.player
		s := next_word
		i := next_integer
			--should error check
		if s.is_equal("ammo") then
			p.add_ammo(i)
		elseif s.is_equal("gems") then
			p.add_gems(i)
		elseif s.is_equal("torches") then
			p.add_torch(i)
		elseif s.is_equal("health") then
			p.add_health(i)
		elseif s.is_equal("score") then
			p.add_score(i)
		end
	end

	command_take is
	local
		s : STRING
		i : INTEGER
		m : STRING
		p : PLAYER
	do
		p := dungeon_master.zzt.player
		s := next_word
		i := next_integer
		m := next_word
			--should error check
		if s.is_equal("ammo") then
			if p.ammo >= i then
				p.add_ammo(-i)
			else
				goto(m)
			end
		elseif s.is_equal("gems") then
			if p.gems >= i then
				p.add_gems(-i)
			else
				goto(m)
			end
		elseif s.is_equal("torches") then
			if p.torch >= i then
				p.add_torch(-i)
			else
				goto(m)
			end
		elseif s.is_equal("health") then
			if p.health >= i then
				p.add_health(-i)
			else
				goto(m)
			end
		elseif s.is_equal("score") then
			if p.score >= i then
				p.add_score(-i)
			else
				goto(m)
			end
		end
	end

	command_set is
	do
		flag_table.put(true, next_word)
	end

	command_clear is
	do
		flag_table.remove(next_word)
	end

	command_char is
	local
		s : STRING
		i : INTEGER
	do
		s := next_word
		if s.is_integer then
			i := s.to_integer
			if i >= 0 and i <= 255 then
				put_chr(i)
			end
		end
	end

	command_cycle is
	local
		s : STRING
		i : INTEGER
	do
		s := next_word
		if s.is_integer then
			i := s.to_integer
			if i >= 1 and i <= 255 then
				cycle := i
			else
				cycle := 3
			end
		else
			cycle := 3
		end
	end

	command_lock is
	do
		locked := true
	end

	command_unlock is
	do
		locked := false
	end

	command_send is
	local
		m : STRING
	do
		m := next_word
		if program @ current_offset = ':' then
			send_message(m)
		else
			goto(m)
		end
	end
	
	send_message(target : STRING) is
	local
		i : INTEGER
		global : BOOLEAN
		list : LINKED_LIST[OBJECT]
		m : STRING
	do
		current_offset := current_offset + 1
		m := next_word
		obit.set_name(target)
		from obit.start
		until obit.is_off
		loop
			if not obit.item.locked then
				obit.item.goto(m)
			end
			obit.next
		end
	end

	command_end is
	do
		current_offset := program.count
	end
	
	command_restart is
	do
		old_offset := 1
		go_forward := false
		time_up := true
	end

	command_go is
	local
		d : INTEGER
	do
		d := get_direction
		if d >= min_dir and d <= max_dir then
			go_forward := try_start_push(d)
		end
		time_up := true
	end

	command_try is
	local
		d : INTEGER
		success : BOOLEAN
	do
		d := get_direction
		if d >= min_dir and d <= max_dir then
			success := try_start_push(d)
		end
		if not success then
			goto(next_word)
		end
		time_up := true
	end

	command_idle is
	do
		time_up := true
	end

	command_walk is
	local
		d : INTEGER
	do
		walk_aim := get_direction
	end

	command_die is
	do
		die
		command_end
		time_up := true
	end

	command_bind is
	local
		i : INTEGER
		target : STRING
		found : BOOLEAN
		list : LINKED_LIST[OBJECT]
		ob : OBJECT
	do
		target := next_word
		list := dungeon_master.zzt.board.object_list
		from i := list.lower
		until found or else i > list.upper
		loop
			ob := list.item(i)
			if ob.id_lower.is_equal(target) then
				found := true
			else
				i := i + 1
			end
		end
		if found then
			if is_bound then
				io.put_string("Uh oh: already bound!%N")
				die_with_code(1)
			end
			is_bound := true
			io.put_string("Binding...%N")
			program := ob.program
			id := ob.id
			id_lower := ob.id_lower
			old_offset := 1
			go_forward := false
		end
		time_up := true
	end

	command_if is
	do
		if get_flag then
			--more undocumented syntax
			--you can have any statement after the if
			io.put_string("if test: true%N")
			skip_space
			old_offset := current_offset
			if next_word.is_equal("then") then
				skip_space
			else
				current_offset := old_offset
			end

			process_command
			old_offset := current_offset
			go_forward := false
				--need this so it doesn't skip a line.
		end
	end

	command_zap is
	local
		target : STRING
		m : STRING
	do
		target := next_word
		if program @ current_offset = ':' then
			-- undocumented(?): can zap labels in other objects
			current_offset := current_offset + 1
			m := next_word

			obit.set_name(target)
			from obit.start
			until obit.is_off
			loop
				if not obit.item.is_bound then
					obit.item.zap(m)
				end
				obit.next
			end
		else
			zap(target)
		end
	end

	zap(m : STRING) is
	local
		i : INTEGER
	do
		i := find_label(m)
		if i <= program.count then
			program.put('%'', i)
		else
			io.put_string("ERR: Bad #zap: ")
			io.put_string(m)
			io.put_string("%N")
			--die_with_code(1)
		end
	end

	command_restore is
	local
		target : STRING
		m : STRING
	do
		target := next_word
		if program @ current_offset = ':' then
			-- undocumented(?): can restore labels in other objects
			current_offset := current_offset + 1
			m := next_word

			obit.set_name(target)
			from obit.start
			until obit.is_off
			loop
				if not obit.item.is_bound then
					obit.item.restore(m)
				end
				obit.next
			end
		else
			restore(target)
		end
	end

	restore(m : STRING) is
	local
		i : INTEGER
		done : BOOLEAN
	do
		from
		until done
		loop
			i := find_char_string('%'', m)
			if i <= program.count then
				program.put(':', i)
			else
				done := true
			end
		end
	end

	oldrestore(m : STRING) is
	local
		i : INTEGER
	do
		i := find_char_string('%'', m)
		if i <= program.count then
			program.put(':', i)
		else
			io.put_string("ERR: Bad #restore: ")
			io.put_string(m)
			io.put_string("%N")
			--die_with_code(1)
		end
	end

	command_put is
	local
		d : INTEGER
		s : STRING
		c : INTEGER
		t : THING
		sq : SQUARE
	do
		d := get_direction
		if d >= min_dir and then d <= max_dir then
			get_kind
			s := tuple_s
			c := tuple_i
			if dungeon_master.zzt.world.thing_table.has(s) then
				t := dungeon_master.zzt.world.thing_table.at(s)
				sq := square.adjacent(d)
				if not sq.dominant.kind.is_equal("sentinel") then
					io.put_string("c is: ")
					io.put_integer(c)
					io.put_string("%N")
					if c = -1 then
						c := sq.dominant.colour_code
					end
					c := t.colour_code_to_colour(c)
					io.put_string("c has become: ")
					io.put_integer(c)
					io.put_string("%N")
					if try_stay_push(d) then
						t := t.instantiate(sq, c)
					else
						if sq.dominant.kind.is_equal(s) then
							sq.dominant.put_colour(c)
						else
							--replace
							sq.clear
							t.add_to_map(sq.map, sq, c)
						end
					end
				end
			else
				--XXX #put: no such object
				io.put_string("no such object: ")
				io.put_string(s)
				io.put_string("%N")
			end
		else
			--XXX #put: bad direction
			io.put_string("bad direction: ")
			io.put_integer(d)
			io.put_string("%N")
		end
	end

	command_become is
	local
		s : STRING
		c : INTEGER
		t : THING
		sq : SQUARE
	do
		get_kind
		s := tuple_s
		if dungeon_master.zzt.world.thing_table.has(s) then
			t := dungeon_master.zzt.world.thing_table.at(s)
			sq := square
			die
			if tuple_i = -1 then
				c := colour
			else
				c := t.colour_code_to_colour(tuple_i)
			end
			t := t.instantiate(sq, c)
			time_up := true
			--goto end as well?
		else
			--XXX #become: no such object
			io.put_string("no such object: ")
			io.put_string(s)
			io.put_string("%N")
		end
	end

	command_change is
	local
		s1, s2 : STRING
		c1, c2 : INTEGER
	do
		get_kind
		s1 := tuple_s
		c1 := tuple_i
		get_kind
		s2 := tuple_s
		c2 := tuple_i
		if dungeon_master.zzt.world.thing_table.has(s1)
			and then dungeon_master.zzt.world.thing_table.has(s2) then
				dungeon_master.zzt.board.change(c1, s1, c2, s2)
		else
			--XXX #change: no such object
			io.put_string("no such object(s): ")
			io.put_string(s1)
			io.put_string(", ")
			io.put_string(s2)
			io.put_string("%N")
		end
	end

	command_endgame is
	do
		dungeon_master.zzt.player.put_health(0)
		time_up := true
	end

	command_shoot is
	local
		b : BULLET
		d : INTEGER
	do
		d := get_direction
		if d >= min_dir and d <= max_dir then
			!!b.make_shoot(square, d, 1)
		end
		time_up := true
	end

	command_throwstar is
	local
		s : STAR
		d : INTEGER
	do
		d := get_direction
		if d >= min_dir and d <= max_dir then
			!!s.make_shoot(square, d, 1)
		end
		time_up := true
	end

	command_play is
	do
		--sound will come much later
	end

	--can't be bothered making a two-tuple
	--integer and string object, so I'll make do with this
	tuple_i : INTEGER
	tuple_s : STRING
	get_kind is
	do
		tuple_s := next_word
		if colour_table.has(tuple_s) then
			tuple_i := colour_table.at(tuple_s)
			tuple_s := next_word
		else
			tuple_i := -1
		end
	end

	colour_table : DICTIONARY[INTEGER, STRING] is
	once
		!!Result.make
		Result.put(1, "blue")
		Result.put(2, "green")
		Result.put(3, "cyan")
		Result.put(4, "red")
		Result.put(5, "purple")
		Result.put(6, "yellow")
		Result.put(7, "white")
	end

	next_line is
	do
		next_cr
		current_offset := current_offset + 1
	end

	next_cr is
	do
		from
		until program.item(current_offset) = '%R'
		loop
			current_offset := current_offset + 1
		end
	end

	skip_space is
	do
		from
		until program.item(current_offset) /= ' '
		loop
			current_offset := current_offset + 1
		end
	end

	next_word : STRING is
	do
		!!Result.make(0)
		skip_space
		from
		until not program.item(current_offset).is_letter_or_digit
		loop
			Result.add_last(program.item(current_offset).to_lower)
			current_offset := current_offset + 1
		end
	end

	next_integer : INTEGER is
	local
		s : STRING
	do
		s := next_word
		if s.is_integer then
			Result := s.to_integer
		end
	end

	goto(label : STRING) is
		--if label doesn't exist, don't do anything
	local
		i : INTEGER
	do
		i := find_label(label)
		if i <= program.count then
			current_offset := i
			-- have to be a bit picky about this
			-- to get it to behave exactly like original
			next_cr
			old_offset := current_offset
			go_forward := false
			-- stop infinite loops
			goto_count := goto_count + 1
			if goto_count = goto_max then
				time_up := true
			end
		end
	end

	find_label(label : STRING) : INTEGER is
	do
		Result := find_char_string(':', label)
	end

	find_char_string(c : CHARACTER; label : STRING) : INTEGER is
		-- return program.count + 1 if not in program
	local
		i : INTEGER
		j : INTEGER
		s : STRING
		found : BOOLEAN
	do
		--search for label
		from i := 1
		until found or else i > program.count
		loop
			if program @ i = c then
				--XXX: inefficient search
				from
					!!s.make(0)
					j := i + 1
				until not (program @ j).is_letter_or_digit
				loop
					s.add_last((program @ j).to_lower)
					j := j + 1
				end
				if s.is_equal(label) then
					found := true
				else
					i := next_line_search(i)
				end
			else
				i := next_line_search(i)
			end
		end
		Result := i
	end

	next_line_search(n : INTEGER) : INTEGER is
	do
		from Result := n
		until program.item(Result) = '%R'
		loop
			Result := Result + 1
		end
		Result := Result + 1
	end

	shot(b : BULLET) is
	do
		if not locked then
			goto("shot")
		end
	end

	blast is
	do
		if not locked then
			goto("blast")
		end
	end

	player_touch(p : PLAYER) is
	do
			io.put_string("Object (")
			io.put_integer(square.location.x)
			io.put_string(", ")
			io.put_integer(square.location.y)
			io.put_string(") [")
			io.put_string(id)
			if is_bound then
				io.put_string(" (B)")
			end
			io.put_string("] ")
			io.put_integer(current_offset)
			io.put_string("/")
			io.put_integer(program.count)
			io.put_string(" touched")
			if locked then
				io.put_string(" but is locked%N")
			else
				io.put_string("%N")
			end
			--if id_lower.is_equal("dark") then
				--io.put_string("special:%N")
				--show_program
			--end
		if not locked then
			goto("touch")
		end
	end

	die is
	local
		list : LINKED_LIST[OBJECT]
	do
		old_die
		walk_aim := direction_idle
		list := square.map.object_list
		io.put_string("removing object from list...%N")
		list.remove(list.index_of(Current))
		io.put_string("done.%N")
	end

	kind : STRING is
	once
		Result := "object"
	end

	zzt_code : INTEGER is
	once
		Result := 36
	end

	is_bound : BOOLEAN

feature {NONE}

	get_direction : INTEGER is
	do
		Result := string_to_direction(next_word)
		inspect Result
		when direction_flow then
			Result := walk_aim
		when direction_opp then
			Result := opposite_direction(get_direction)
		when direction_cw then
			Result := clockwise_direction(get_direction)
		when direction_ccw then
			Result := counterclockwise_direction(get_direction)
		when direction_rndns then
			rand.next
			if rand.last_real < 0.5 then
				Result := north
			else
				Result := south
			end
		when direction_rndne then
			rand.next
			if rand.last_real < 0.5 then
				Result := north
			else
				Result := east
			end
		when direction_rndp then
			rand.next
			if rand.last_real < 0.5 then
				Result := clockwise_direction(get_direction)
			else
				Result := counterclockwise_direction(get_direction)
			end
		when direction_seek then
			Result := seek_direction(square.location, dungeon_master.player.square.location)
		else
			--unimplemented direction or standard direction (n, s, e, w, i)
			if Result > max_dir then
				io.put_string("Uh oh: unimplemented direction: ")
				io.put_integer(Result)
				io.put_string("%N")
				die_with_code(1)
			end
		end
	end

	get_flag : BOOLEAN is
	local
		flag : STRING
		d : INTEGER
		dx, dy : INTEGER
	do
		flag := next_word
		if flag.is_equal("blocked") then
			Result := true
			d := get_direction
			if d >= min_dir and d <= max_dir then
				if can_walk(d) then
					Result := false
				end
			end
		elseif flag.is_equal("alligned") then
			--annoying spelling mistake
			Result :=
				dungeon_master.player.square.location.x = square.location.x
				or else dungeon_master.player.square.location.y = square.location.y
		elseif flag.is_equal("contact") then
			dx := dungeon_master.player.square.location.x - square.location.x
			dy := dungeon_master.player.square.location.y - square.location.y
			Result := dx * dy = 0 and then dx + dy = 1
		elseif flag.is_equal("not") then
			--Undocumented
			Result := not get_flag
		elseif flag.is_equal("any") then
			--Undocumented
			get_kind
			Result := square.map.find_first(tuple_i, tuple_s) /= Void
		else
			Result := flag_table.has(flag)
		end
	end

	flag_table : DICTIONARY[BOOLEAN, STRING] is
	once
		!!Result.make
	end

	command_table : DICTIONARY[OBJECT_COMMAND, STRING] is
	local
		oc : OBJECT_COMMAND
	once
		!!Result.make
		!OC_GO!oc.make
		Result.put(oc, "go")
		!OC_TRY!oc.make
		Result.put(oc, "try")
		!OC_WALK!oc.make
		Result.put(oc, "walk")
		!OC_END!oc.make
		Result.put(oc, "end")
		!OC_IDLE!oc.make
		Result.put(oc, "idle")
		!OC_SEND!oc.make
		Result.put(oc, "send")
		!OC_DIE!oc.make
		Result.put(oc, "die")
		!OC_UNLOCK!oc.make
		Result.put(oc, "unlock")
		!OC_LOCK!oc.make
		Result.put(oc, "lock")
		!OC_RESTART!oc.make
		Result.put(oc, "restart")
		!OC_IF!oc.make
		Result.put(oc, "if")
		!OC_SET!oc.make
		Result.put(oc, "set")
		!OC_CLEAR!oc.make
		Result.put(oc, "clear")
		!OC_CHAR!oc.make
		Result.put(oc, "char")
		!OC_CYCLE!oc.make
		Result.put(oc, "cycle")
		!OC_GIVE!oc.make
		Result.put(oc, "give")
		!OC_TAKE!oc.make
		Result.put(oc, "take")
		!OC_ZAP!oc.make
		Result.put(oc, "zap")
		!OC_RESTORE!oc.make
		Result.put(oc, "restore")
		!OC_BIND!oc.make
		Result.put(oc, "bind")
		!OC_PUT!oc.make
		Result.put(oc, "put")
		!OC_BECOME!oc.make
		Result.put(oc, "become")
		!OC_PLAY!oc.make
		Result.put(oc, "play")
		!OC_CHANGE!oc.make
		Result.put(oc, "change")
		!OC_ENDGAME!oc.make
		Result.put(oc, "endgame")
		!OC_SHOOT!oc.make
		Result.put(oc, "shoot")
		!OC_THROWSTAR!oc.make
		Result.put(oc, "throwstar")
	end

	obit : ITERATOR_ON_OBJECT is
	once
		!!Result.make
	end
end
